#import "BLKFunction.h"

@implementation BLKFunction

#pragma mark - Return object to stop an enumeration
const id BLKStopProcessing = @"BLKStopProcessing";

#pragma mark - Standalone operations

+ (NSMutableArray *)shuffleArray:(NSArray *)array {
    id mutableArray = [array mutableCopy];
    // Fisher-Yates Shuffle
    for (int i = (int) [mutableArray count] - 1; i > 0; i--) {
        [mutableArray exchangeObjectAtIndex:arc4random_uniform(i + 1) withObjectAtIndex:i];
    }
    return mutableArray;
}

#pragma mark - Operations defined in terms of public fold method

+ (NSMutableArray *)takeFromArray:(NSArray *)array numberOfItems:(NSUInteger)nrToTake {
    return [BLKFunction foldArray:array toAccumulator:@[]
                 withIndexedBlock:^(id acc, id obj, NSUInteger idx) {
                     if (idx >= nrToTake) return BLKStopProcessing;
                     else [acc addObject:obj];
                     return acc;
                 }];
}

+ (NSMutableArray *)dropFromArray:(NSArray *)array numberOfItems:(NSUInteger)nrToDrop {
    return [BLKFunction foldArray:array toAccumulator:@[]
                 withIndexedBlock:^(id acc, id obj, NSUInteger idx) {
                     if (idx >= nrToDrop) [acc addObject:obj];
                     return acc;
                 }];
}

+ (NSMutableArray *)filterArray:(NSArray *)array withBlock:(BLKTypePred1)block {
    return [BLKFunction foldArray:array toAccumulator:@[] withBlock:^(id acc, id obj) {
        if (block(obj)) [acc addObject:obj];
        return acc;
    }];
}

+ (id)reduceArray:(NSArray *)array withBlock:(BLKTypeFn2)block {
    id first = [BLKFunction ios6FirstObject:array];
    return [BLKFunction foldArray:[BLKFunction dropFromArray:array numberOfItems:1]
                    toAccumulator:first withBlock:block];
}

+ (NSMutableArray *)reverseArray:(NSArray *)array {
    return [BLKFunction foldArray:array usingOptions:NSEnumerationReverse toAccumulator:@[]
                 withIndexedBlock:^(id acc, id obj, NSUInteger idx) {
                     [acc addObject:obj];
                     return acc;
                 }];
}

#pragma mark - Operations defined in terms of private fold method

+ (NSMutableArray *)mapArray:(NSArray *)array withBlock:(BLKTypeFn1)block {
    return [BLKFunction foldSeq:array usingOptions:0 into:@[] viaFn1:block fn1Idx:nil fn2:nil fn2Idx:nil];
}

+ (NSMutableArray *)mapArray:(NSArray *)array withIndexedBlock:(BLKTypeFn1Idx)block {
    return [BLKFunction foldSeq:array usingOptions:0 into:@[] viaFn1:nil fn1Idx:block fn2:nil fn2Idx:nil];
}

+ (id)foldArray:(NSArray *)array toAccumulator:(id)accumulator withBlock:(BLKTypeFn2)block {
    return [BLKFunction foldSeq:array usingOptions:0 into:accumulator viaFn1:nil fn1Idx:nil fn2:block fn2Idx:nil];
}

+ (id)foldArray:(NSArray *)array toAccumulator:(id)accumulator withIndexedBlock:(BLKTypeFn2Idx)block {
    return [BLKFunction foldSeq:array usingOptions:0 into:accumulator viaFn1:nil fn1Idx:nil fn2:nil fn2Idx:block];
}

+ (id)foldArray:(NSArray *)array usingOptions:(NSEnumerationOptions)options
  toAccumulator:(id)accumulator withIndexedBlock:(BLKTypeFn2Idx)block {
    return [BLKFunction foldSeq:array usingOptions:options into:accumulator
                         viaFn1:nil fn1Idx:nil fn2:nil fn2Idx:block];
}

#pragma mark - Dictionary

+ (id)foldDictionary:(NSDictionary *)dict toAccumulator:(id)accumulator withBlock:(BLKTypeFn3)block {
    return [BLKFunction foldDictionary:dict usingOptions:0 into:accumulator viaFn3:block fn3Idx:nil];
}

+ (id)foldDictionary:(NSDictionary *)dict toAccumulator:(id)accumulator
    withIndexedBlock:(BLKTypeFn3Idx)block {
    return [BLKFunction foldDictionary:dict usingOptions:0 into:accumulator viaFn3:nil fn3Idx:block];
}

+ (id)foldDictionary:(NSDictionary *)dict usingOptions:(NSEnumerationOptions)options
       toAccumulator:(id)accumulator withIndexedBlock:(BLKTypeFn3Idx)block {
    return [BLKFunction foldDictionary:dict usingOptions:options into:accumulator viaFn3:nil fn3Idx:block];
}

#pragma mark - Conjoin two or more objects or collections

+ (id)conjoin:(id)cage1 and:(id)cage2 {
    return [BLKFunction conjoin:cage1 and:cage2 interpose:nil interleave:nil];
}

// conjoin multiple objects
+ (id)conjoin:(id)cage1 and:(id)cage2 and:(id)cage3 {
    return [BLKFunction conjoin:[BLKFunction conjoin:cage1 and:cage2] and:cage3];
}

+ (id)conjoin:(id)cage1 and:(id)cage2 interpose:(id)theInterpose {
    return [BLKFunction conjoin:cage1 and:cage2 interpose:theInterpose interleave:nil];
}

+ (id)conjoin:(id)cage1 and:(id)cage2 interpose:(id)theInterpose end:(id)theEnd {
    return [BLKFunction conjoin:[BLKFunction conjoin:cage1 and:cage2 interpose:theInterpose] and:theEnd];
}

+ (id)conjoin:(id)cage1 and:(id)cage2 interleave:(NSArray *)theInterleave {
    return [BLKFunction conjoin:cage1 and:cage2 interpose:nil interleave:theInterleave];
}

#pragma mark - Private methods

+ (id)conjoin:(id)cage1 and:(id)cage2
    interpose:(id)theInterpose interleave:(NSArray *)theInterleave {
    if (!cage1) return nil;
    id zoo = [BLKFunction mutableCopyIfImmutable:cage1];
    if (!cage2) return zoo;

    BLKTypeFn2 conjoiner = [BLKFunction conjoinerForZoo:zoo];

    // single object join>
    if (![cage2 conformsToProtocol:@protocol(NSFastEnumeration)]) return conjoiner(zoo, cage2);

    BOOL zooIsString = [zoo isKindOfClass:[NSString class]];
    BOOL cage2IsDictionary = [cage2 isKindOfClass:[NSDictionary class]];
    if ([zoo isKindOfClass:[NSDictionary class]]) { // todo: how to handle this?
        if (cage2IsDictionary) {  // no point iterating, just return
            [zoo addEntriesFromDictionary:cage2];
            return zoo;
        }
        theInterleave = nil;
        theInterpose = nil;
    }

    NSUInteger idx = 0;
    for (id obj in cage2) {
        id animal = obj;
        if (cage2IsDictionary) {
            id key = obj;
            id value = cage2[obj];
            if (zooIsString) animal = [NSString stringWithFormat:@"{%@:%@}", key, value]; //todo: what to do?
            else animal = @[key, value];
        }
        if (theInterleave && (idx >= [theInterleave count])) break;
        if (theInterpose && idx) zoo = conjoiner(zoo, theInterpose);

        zoo = conjoiner(zoo, animal);
        if (theInterleave) zoo = conjoiner(zoo, theInterleave[idx]);
        idx++;
    }
    return zoo;
}

// ios6 hides firstObject
+ (id)ios6FirstObject:(NSArray *)array {
    if (array == nil) return nil;
    else if (array.count == 0) return nil;
    else return array[0];
}

+ (id)mutableCopyIfImmutable:(id)obj {
    // mutable objects
    if ([obj isMemberOfClass:[NSString class]]) return [obj mutableCopy];

    // and collections
    if ([obj conformsToProtocol:@protocol(NSFastEnumeration)]
            && ![obj respondsToSelector:@selector(addObject:)]) {
        return [obj mutableCopy];
    }
    return obj;
}

+ (id)foldSeq:(id)sequence usingOptions:(NSEnumerationOptions)options into:(id)accumulator
       viaFn1:(BLKTypeFn1)fn1 fn1Idx:(BLKTypeFn1Idx)fn1Idx
          fn2:(BLKTypeFn2)fn2 fn2Idx:(BLKTypeFn2Idx)fn2Idx {
    // "maps" need to add to the accumulator
    BOOL addToAccumulator = (fn1 || fn1Idx) ? YES : NO;

    // local accumulator
    __block id acc = [BLKFunction mutableCopyIfImmutable:accumulator ? accumulator : @[]];

    [sequence enumerateObjectsWithOptions:options usingBlock:^(id obj, NSUInteger idx, BOOL *stop) {
        // which function to run
        id ret;
        if (fn1) ret = fn1(obj);
        else if (fn1Idx) ret = fn1Idx(obj, idx);
        else if (fn2) ret = fn2(acc, obj);
        else if (fn2Idx) ret = fn2Idx(acc, obj, idx);
        // check if the user has indicate to break out of the enumeration
        if (ret == BLKStopProcessing) *stop = YES;
        else if (addToAccumulator) [acc addObject:ret];
        else acc = ret;
    }];

    return acc;
}

+ (id)foldDictionary:(NSDictionary *)dict usingOptions:(NSEnumerationOptions)options into:(id)accumulator
              viaFn3:(BLKTypeFn3)fn3 fn3Idx:(BLKTypeFn3Idx)fn3Idx {
    // local accumulator
    __block id acc = [BLKFunction mutableCopyIfImmutable:accumulator ? accumulator : @[]];

    __block NSUInteger idx = 0;
    [dict enumerateKeysAndObjectsWithOptions:options usingBlock:^(id obj1, id obj2, BOOL *stop) {
        // which function to run
        id ret;
        if (fn3) ret = fn3(acc, obj1, obj2);
        else if (fn3Idx) {
            ret = fn3Idx(acc, obj1, obj2, idx);
            idx++;
        }
        // check if the user has indicate to break out of the enumeration
        if (ret == BLKStopProcessing) *stop = YES;
        else acc = ret;
    }];

    return acc;
}

/**
* What to do when the exampleCage is a dictionary, converted to [[key, value]]. This is ok for now.
*/
+ (BLKTypeFn2)conjoinerForZoo:(id)zoo {
    __block id key;
    if ([zoo isKindOfClass:[NSString class]]) {
        return ^(NSString *acc, id obj1) {
            if ([obj1 isKindOfClass:[NSString class]]) return [acc stringByAppendingString:obj1];
            return [acc stringByAppendingString:[obj1 description]];
        };
    } else if ([zoo isKindOfClass:[NSNumber class]]) {
        return ^(NSNumber *acc, id obj1) {
            if ([obj1 isKindOfClass:[NSNumber class]]) return @(acc.doubleValue + [obj1 doubleValue]);
            else return acc;
        };
    } else if ([zoo isKindOfClass:[NSDictionary class]]) {
        return ^(NSMutableDictionary *acc, id obj1) {
            if (key) {
                acc[key] = obj1;
                key = nil;
            } else if ([obj1 isKindOfClass:[NSArray class]] && [obj1 count] == 2) {
                id key = [BLKFunction ios6FirstObject:obj1];
                acc[key] = [obj1 lastObject];
            } else {
                key = obj1;
            }
            return acc;
        };
    } else if ([zoo conformsToProtocol:@protocol(NSFastEnumeration)]) {
        return ^(id acc, id obj1) {
            [acc addObject:obj1];
            return acc;
        };
    }
    return ^(id p1, id p2) {
        return [@"" stringByAppendingFormat:@"%@%@", p1, p2];
    };
}

@end